/*
 * FluxVideo.cpp
 *
 *  Created on: Jun 6, 2013
 *      Author: exploit
 *
 * Classe auquel on donne un url et permet d'afficher le flux vidéo associé.
 * L'affichage se fait à l'aide d'openCV. Converti le flux à l'aide de FFMpeg
 * si néscessaire.
 */

#include "Capture.hpp"

/*******************************************************************************
								CONSTRUCTEUR/DESTRUCTEUR
*******************************************************************************/
static const int64_t timeout= 2;// temps avant que l'on arrête la capture en seconde
static bool timeoutExpire = false;
static int detecteErreurReception(void*);


/***
 * Créé une instance de capture video.
 * @theUrl path vers la video que cette instance va devoir capter.
 */
Capture::Capture(const std::string &theUrl)
		: url(theUrl), correct(true), fluxFFmpeg(NULL),
		  packetEncode(NULL), frameDecode(NULL),
		  streamVideo(NULL)
{
	init();
}

Capture::Capture(const Capture & theFlux)
		: url(theFlux.url), correct(theFlux.correct),
		fluxFFmpeg(theFlux.fluxFFmpeg), packetEncode(theFlux.packetEncode),
		frameDecode(theFlux.frameDecode),
		streamVideo(theFlux.streamVideo),
		cbp(theFlux.cbp)
{

}

Capture::~Capture()
{
	avcodec_close(streamVideo->codec);
	avformat_free_context(fluxFFmpeg);
	av_free(frameDecode);
	av_free_packet(packetEncode);
	av_free(streamVideo->codec);
	//swr_free(contextTraitementImage);

	delete cbp;
}

/*******************************************************************************
								ACCESSEURS
*******************************************************************************/

std::string Capture::getUrl() const
{
	return url;
}

/*
 * Retourne la largueur (nombre de colonne) de l'image.
 */
int Capture::getLargeur()const
{
	return streamVideo->codec->width;
}

/*
 * Retourne la hauteur (nombre de ligne) de l'image.
 */
int Capture::getHauteur()const
{
	return streamVideo->codec->height;
}

enum PixelFormat Capture::getFormat()const
{
	return streamVideo->codec->pix_fmt;
}

/***
 * Retourne les fps ou framerate de la vidéo.
 */
double Capture::getFPS() const
{
	AVRational fraction = streamVideo->r_frame_rate;
	return fraction.num / fraction.den;
}

int Capture::getFramSize() const
{
	return streamVideo->codec->height * streamVideo->codec->width;
}

/*
 * Retourne le flux au format FFMpeg
 */
AVFormatContext* Capture::getFluxFFmpeg()const
{
	return fluxFFmpeg;
}

/*
 * Retourne la valeur de l'attribut frameEncode.
 */
AVPacket* Capture::getDataEncode()const
{
	return packetEncode;
}

/*
 * Retourne la valeur de l'attribut frameDecode.
 */
AVFrame* Capture::getDataDecode()const
{
	return frameDecode;
}

/*
 * Retourne les paramètres associées à l'extraction du flux vidéo.
 */
AVCodecContext* Capture::getDecodeContext() const
{
	return streamVideo->codec;
}

bool Capture::estCorrect() const
{
	return correct && !timeoutExpire;
}

/*******************************************************************************
								FONCTIONS PUBLIQUES
									TRAITEMENTS
*******************************************************************************/

/**
 * Fonction qui va lire la prochaine image sur le flux.
 * Les images lus seront accessible via getFrameDecode pour l'image
 * au format FFmpeg et getDataRGB au format OpenCV.
 * return : vrai si une nouvelle image a été lue.
 */
bool Capture::lire()
{
	/**
	 * On prépare le call back afin qu'il arrête la fonction si la réception est
	 * interompue.
	 */
	cbp->actif = true;
	cbp->start = av_gettime();

	//On décode la vidéo
	return lirePacket();
}

/*******************************************************************************
								FONCTIONS PROTECTED
									INITIALISATION
*******************************************************************************/
void Capture::init()
{
	initFluxFFmpeg();
	initStreamVideo();
	initAlloueAttribut();
}


/*
 * Fonction qui capte et analyse le flux FFmpeg associé à l'url.
 */
void Capture::initFluxFFmpeg()
{
	//TODO initialisation autre part
	av_register_all();
	avcodec_register_all();
	//avcodec_init();//serait dévalué
	avformat_network_init();

	if(fluxFFmpeg == NULL)
		fluxFFmpeg = avformat_alloc_context();

	/*
	 * On ajoute une fonction de callback pour éviter que le processus ne bloque
	 * le programme si la réception ne fonctionne pas
	 */
	cbp = new CallBackParam(av_gettime(), true, timeout);
	AVIOInterruptCB cb = {detecteErreurReception, cbp};
	fluxFFmpeg->interrupt_callback = cb;

	//Ouvre le flux
	if(avformat_open_input(&fluxFFmpeg, url.c_str(), NULL, NULL) < 0){
		fprintf(stderr, "Capture::initFluxFFmpeg : "
				"Impossible d'ouvrir le fichier %s.\n", url.c_str());
		correct = false;
		return;
	}

	/*
	 * Pour que la fonction de callback laisse le temps à la capture d'initialiser
	 * la vidéo
	 */
	cbp->actif = false;

	//Extrait les informations du flux. TODO Provoque une fuite mémoire
	if(avformat_find_stream_info(fluxFFmpeg, NULL) < 0){
		fprintf(stderr, "Capture::initFluxFFmpeg : "
				"Could not find stream information.\n");
		correct = false;
		return;
	}

	if(fluxFFmpeg->nb_streams == 0){
		fprintf(stderr, "Capture::initFluxFFmpeg : "
				"Aucun stream associé à cet url.\n");
		correct = false;
		return;
	}
//	ffmpegTools::print_info(fluxFFmpeg);
//
//	printf("###FIN Capture::initFluxFFmpeg###\n");
}

/*
 * Initialise le flux vidéo ainsi que les codecs associés.
 */
void Capture::initStreamVideo()
{
	if(correct == false || fluxFFmpeg == NULL){
//		fprintf(stderr, "Capture::initStreamVideo : "
//				"Le fichier n'est pas correct.\n");
		return;
	}
	//Numéro du flux que l'on va séléctionner parmi tous les flux de fmt_ctx
	int numFlux;

	//On cherche le flux de fluxFFmpeg qui correspond au type
	numFlux = av_find_best_stream(fluxFFmpeg, AVMEDIA_TYPE_VIDEO, -1, -1, NULL, 0);
	if (numFlux < 0) {
		if(numFlux == (int) AVERROR_STREAM_NOT_FOUND)
			fprintf(stderr, "Capture::initStreamVideo : "
					"Aucun flux vidéo n'existe dans ce stream : %s.\n",
					url.c_str());
		if(numFlux == (int) AVERROR_DECODER_NOT_FOUND)
			fprintf(stderr, "Capture::initStreamVideo : "
					"Aucun décodeur n'a été trouvé pour ce stream : %s.\n",
					url.c_str());
		correct = false;
		return;
	}
	streamVideo = fluxFFmpeg->streams[numFlux];
	if(streamVideo == NULL){
		fprintf(stderr, "Capture::initStreamVideo : "
				"StreamVideo == NULL.\n");
	}

	AVCodecContext *decodeurContext = NULL;
	AVCodec *dec = NULL;

	//On cherche un décodeur pour le flux
	decodeurContext = streamVideo->codec;
	dec = avcodec_find_decoder(decodeurContext->codec_id);
	if (dec == NULL) {
		fprintf(stderr, "Capture::initStreamVideo : Impossible de trouver le codec.\n");
		correct = false;
		return;
	}

	//On charge le décodeur TODO provoque une fuite mémoire
	if (avcodec_open2(decodeurContext, dec, NULL) < 0) {
		fprintf(stderr, "Capture::initStreamVideo : Impossible d'ouvrir le codec.\n");
		correct = false;
		return;
	}

	//printf("###FIN Capture::initStreamVideo### \n");
}

/*
 * Fonction qui alloue tous les attributs utilisés dans
 * la classe lors des traitements.
 */
void Capture::initAlloueAttribut()
{
	if(!correct)
		return;

	//AVPacket* packetEncode
	packetEncode = (AVPacket*)av_malloc(sizeof(AVPacket));
	av_init_packet(packetEncode);

	//AVFrame* frameDecode
	frameDecode = avcodec_alloc_frame();
	//Pour le streaming vlc, il est important que les données images soient déjà initialisées
	//TODO revoir car pas très propre
//	frameDecode->data = (uint8_t*)malloc(sizeof(uint8_t*));
//	frameDecode->data[0] = (uint8_t*)malloc(sizeof(uint8_t*));
}

/*******************************************************************************
								FONCTIONS PROTECTED
									TRAITEMENT
*******************************************************************************/

/***
 * Décode la vidéo et met à jour les champs packet encode et
 * frameDecode.
 * @return vrai si la lecture et la decompression du packet s'est bien deroulee.
 */
bool Capture::lirePacket()
{
	bool continuer = true;
	while (continuer){
		if(packetEncode != NULL)
			av_free_packet(packetEncode);
		int ret = av_read_frame(fluxFFmpeg, packetEncode);
		if(ret == 1){//en cas de fin de timeout ?
			BOOST_LOG_TRIVIAL(debug) << "Capture::lirePacket "
					"Timeout pour la lecture de frame a expire.";
			return false;
		}else if(ret < 0){
			BOOST_LOG_TRIVIAL(debug) << "Capture::lirePacket "
					"Erreur pendant la lecture de frame.";
			return false;
		}else if(ret > 1 ){
			BOOST_LOG_TRIVIAL(debug) << "Capture::lirePacket "
					"Valeur de retour inconnue.";
			return false;
		}else if(ret == 0){//si la lecture s'est bien deroulee
			//on decode la video
			continuer = !decodeVideo();
		}
	}
	return true;
}

/***
 * Fonction qui va decode le packet packetEncode.
 * @return vrai si la video a bien ete decode, faux dans le cas contraire
 */
bool Capture::decodeVideo()
{
	//boucle de lecture
	if(packetEncode->stream_index == streamVideo->index){
		// Decode video frame
		int frameFinished = 0;
		int in_used = avcodec_decode_video2(streamVideo->codec,
				frameDecode,
				&frameFinished,
				packetEncode);
		//si un frame à été décompréssé
		if(in_used > 0 && frameFinished != 0){
			return true;
		}
		//En cas d'erreur
		if(in_used == 0 || frameFinished == 0){
			fprintf(stderr, "Capture::decodeVideo "
					"Le frame n'a pu être décompréssé\n");
		}
		if (in_used < 0){
			fprintf(stderr, "Capture::decodeVideo "
					"erreur lors du décodage d'un paquet.\n");
		}
	}
	return false;
//	printf("###FIN Capture::decodeVideo### \n");
}

/***
 * Fonction de callback qui sera appelée lors de l'initialisation de AVFormatContext
 * avec la fonction av_open_input. Elle sera aussi appellee lors de av_read_frame.
 * Elle arrête la fonction d'initialisation Lorsque le temps d'exécution de la
 * fonction est > à la variable timeout.
 */
static int detecteErreurReception(void* context)
{
	CallBackParam* ptrCBP	=	(CallBackParam*)context;

	if(!ptrCBP->actif)
		return 0;
	//vizILogger *pLog		=	ptrCapture->get_logger();
	int64_t start			=	ptrCBP->start;

//	printf("start : %lld, gettime : %lld", (long long)start, (long long)av_gettime());
//	printf(" différence : %lld\n", (long long)(av_gettime() - start));

	// timeout after x seconds of no activity
	if ((av_gettime() - start) > (ptrCBP->timeout * 1000000))
	{
		//pLog->Log("<! Timeout : read_interrupt_cb (%i).. (%li -> %li)", timeout/1000, start ,(av_gettime()/1000));
		fprintf(stderr, "detecteErreurReception : "
				"timeout, le flux est interompu.\n");
		timeoutExpire = true;
		return 1;//interrupt
	}
	return 0;
}